/******************************************************************************

   Copyright 1991 Lee Iverson

   Permission to use, copy, modify, and distribute this software and its
   documentation for any purpose without fee is hereby granted without fee,
   provided that the above copyright notice appear in all copies and that
   both that copyright notice and this permission notice appear in
   supporting documentation, and that the name of McGill not be used in
   advertising or publicity pertaining to distribution of the software
   without specific, written prior permission.  McGill makes no
   representations about the suitability of this software for any purpose.
   It is provided "as is" without express or implied warranty.

   MCGILL DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL
   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL MCGILL
   BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
   WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION
   OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
   CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

   Author:
     Lee Iverson <leei@mcrcim.mcgill.edu>,
     McGill Research Centre for Intelligent Machines (McRCIM)

******************************************************************************/

static char RCSid[] = "$Id: pFile.c,v 0.9.1.1 1991/11/06 00:09:25 leei Exp $";

/* 
 * A file abstraction with unlimited pushback.  This is essential for 
 * identifying and reading images from pipes.  Lookahead is also 
 * important for some other ideas and the one-character STDIO limit is 
 * just not workable.
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>

#include "proto.h"
#include "pFile.h"

#define BSIZE (64)
#define HSIZE (8)
#define ISIZE (BSIZE-HSIZE)
#define NSIZE(n) (BSIZE*(((n)+BSIZE-1)/BSIZE)-HSIZE)

static int chartable_init = 0;
static void init_chartable();
_SIGNED char _pfcharTable[256];

pFILE
pfopen _IMPL((fname), _CONST char *fname )
{
    pFILE new;
    FILE *fp = fopen(fname,"r");

    if ( !chartable_init ) init_chartable();
    if ( !fp ) return 0;
    
    new = (pFILE) malloc(sizeof(*new));
    new->fp = fp;
    new->pbuf = (unsigned char *) malloc(ISIZE);
    new->pblen = ISIZE;
    new->pb_l = new->pb_e = new->pbuf + new->pblen; 
    return new;
}

pFILE
pfreopen _IMPL((fp), FILE *fp )
{
    pFILE new = (pFILE) malloc(sizeof(*new));

    if ( !chartable_init ) init_chartable();
    new->fp = fp;
    new->pbuf = (unsigned char *) malloc(ISIZE);
    new->pblen = ISIZE;
    new->pb_l = new->pb_e = new->pbuf + new->pblen; 
    return new;
}

static void
pbrealloc _IMPL((incr, pf, freeit), int incr _AND pFILE pf _AND int freeit )
{
    int newsize = NSIZE(pf->pblen+BSIZE+incr);
    unsigned char *nbuf = (unsigned char *) malloc(newsize);
    int eoff = pf->pb_l - pf->pb_e;
    
    bcopy(pf->pb_e, nbuf + newsize - eoff, eoff );
    pf->pb_l = nbuf + newsize;
    pf->pb_e = pf->pb_l - eoff;
    pf->pblen = newsize;
    if ( freeit ) free(pf->pbuf);
    pf->pbuf = nbuf;
}

pFILE
pfstropen _IMPL((str, len), char *str _AND int len )
{
    pFILE new = (pFILE) malloc(sizeof(*new));

    if ( !chartable_init ) init_chartable();
    new->fp = NULL;
    new->pb_e = new->pbuf = (unsigned char *) str;
    new->pblen = len;
    new->pb_l = new->pbuf + new->pblen; 
    pbrealloc( 0, new, 0 );
    return new;
}

int
pfclose _IMPL((pf), pFILE pf )
{
    int ret = 0;
    if ( pf->fp ) ret = fclose(pf->fp);
    
    free(pf->pbuf);
    free(pf);
    return ret;
}

int
pfpeek _IMPL((ptr, size, n, pf),
	     void *ptr _AND int size _AND int n _AND pFILE pf )
{
    int ret = pfread(ptr,size,n,pf);

    pfungetm(ptr,size,ret,pf);
    return ret;
}

int
pfpushback _IMPL((buf, pf), char **buf _AND pFILE pf )
{
    if ( buf ) *buf = (char *) pf->pb_e;
    return (pf->pb_l - pf->pb_e);
}

int
pfread _IMPL((ptr, size, n, pf),
	     void *ptr _AND int size _AND int n _AND pFILE pf )
{
    char *p = ptr;
    unsigned int tsize = size * n, nt = 0;

    /* Get data from pushback. */
    if ( pf->pb_e < pf->pb_l ) {
	unsigned int avail = pf->pb_l - pf->pb_e;
	if ( tsize < avail ) avail = tsize;
	bcopy(pf->pb_e, p, avail);
	p += avail; pf->pb_e += avail; nt += avail;
    }

    /* Pushback exhausted.  Try stream. */
    if ( nt < tsize && pf->fp ) {
	int nread = fread(p,1,tsize-nt,pf->fp);
	nt += nread;
    }

    /* Pushback extras past last block. */
    if ( nt%size != 0 ) {
	pfungetm(((char*)ptr) + nt - (nt%size), 1, nt%size, pf);
	nt -= nt%size;
    }
    
    return nt/size;
}

int
pfungetm _IMPL((ptr, size, n, pf),
	       void *ptr _AND int size _AND int n _AND pFILE pf )
{
    unsigned int tsize = size * n;

    /* Reallocate the pushback buffer to make room. */
    if ( tsize > pf->pb_e - pf->pbuf ) {
	pbrealloc(tsize - (pf->pb_e - pf->pbuf),pf,1);
	return pfungetm(ptr,size,n,pf);
    }

    bcopy( ptr, pf->pb_e -= tsize, tsize );
    return n;
}

int
pfungets _IMPL((s, pf), char *s _AND pFILE pf )
{
    int len = strlen(s);
    return pfungetm( s, 1, len, pf );
}

void
pfungetc _IMPL((c, pf), char c _AND pFILE pf )
{
    pfungetm( &c, 1, 1, pf );
}

/* Walk over spaces. */

int
pfnextc _IMPL((pf), pFILE pf )
{
    int c;

    for ( c = pfgetc(pf); c > 0 && isspace(c); c = pfgetc(pf) );
    return c;
}

char *
pfgets _IMPL((s, n, pf), char *s _AND int n _AND pFILE pf )
{
    int c;
    char *p;

    for ( p = s; (c = pfgetc(pf)) > 0 && c != '\n' && n; --n, (p? ++p: 0) )
      if (p) *p = c;
    if ( n && c == '\n' ) --n, (p? (*p++ = c): 0);
    if ( n && c == '\0' ) ++n, pfungetc(c, pf);
    if ( n && p ) *p = '\0';
    return s;
}

char *
pfgetw _IMPL((s, n, pf), char *s _AND int n _AND pFILE pf )
{
    int c;
    char *p;

    if ( (c = pfnextc(pf)) <= 0 ) return NULL;

    /* Now read until a space. */
    for ( p = s, *p++ = c, --n; n && (c=pfgetc(pf)) > 0 && !isspace(c); --n )
      *p++ = c;

    if ( isspace(c) ) pfungetc(c,pf);
    if ( n ) *p = '\0';

    return s;
}

int
pfgetlint _IMPL((val, pf), long *val _AND pFILE pf )
{
    int c;
    _SIGNED char sgn = 1;
    long accum = 0;

    /* Walk over spaces. */
    if ( (c = pfnextc(pf)) <= 0 ) return 0;

    /* Now read until a non-integer value. */
    if ( c == '-' ) sgn = -1, c = pfgetc(pf);
    else if ( c == '+' ) sgn = 1, c = pfgetc(pf);

    if ( !isdigit(c) ) return 0;
    for ( ; c > 0 && isdigit(c); c = pfgetc(pf) )
      accum = 10*accum + pfcharval(c);
    if ( c >= 0 ) pfungetc(c,pf);

    *val = sgn*accum;
    return 1;
}

int
pfgetulint _IMPL((val, pf), unsigned long *val _AND pFILE pf )
{
    int c;
    unsigned long accum = 0;

    /* Walk over spaces. */
    if ( (c = pfnextc(pf)) <= 0 ) return 0;

    /* Now read until a non-integer value. */
    if ( !isdigit(c) ) return 0;
    for ( ; c > 0 && isdigit(c); c = pfgetc(pf) )
      accum = 10*accum + pfcharval(c);
    if ( c >= 0 ) pfungetc(c,pf);

    *val = accum;
    return 1;
}

int
pfgetint _IMPL((val, pf), int *val _AND pFILE pf )
{
    long lval;
    int ret = pfgetlint(&lval,pf);
    *val = lval;
    return ret;
}

int
pfgetuint _IMPL((val, pf), unsigned int *val _AND pFILE pf )
{
    unsigned long lval;
    int ret = pfgetulint(&lval,pf);
    *val = lval;
    return ret;
}

int
pfgethint _IMPL((val, pf), short *val _AND pFILE pf )
{
    long lval;
    int ret = pfgetlint(&lval,pf);
    *val = lval;
    return ret;
}

int
pfgetuhint _IMPL((val, pf), unsigned short *val _AND pFILE pf )
{
    unsigned long lval;
    int ret = pfgetulint(&lval,pf);
    *val = lval;
    return ret;
}

int
pfgetcint _IMPL((val, pf), char *val _AND pFILE pf )
{
    long lval;
    int ret = pfgetlint(&lval,pf);
    *val = lval;
    return ret;
}

int
pfgetucint _IMPL((val, pf), unsigned char *val _AND pFILE pf )
{
    unsigned long lval;
    int ret = pfgetulint(&lval,pf);
    *val = lval;
    return ret;
}


static void
init_chartable _IMPL_VOID
{
    int i;
    for ( i = 0; i < 256; ++i ) _pfcharTable[i] = -1;
    _pfcharTable['0'] = 0;
    _pfcharTable['1'] = 1;
    _pfcharTable['2'] = 2;
    _pfcharTable['3'] = 3;
    _pfcharTable['4'] = 4;
    _pfcharTable['5'] = 5;
    _pfcharTable['6'] = 6;
    _pfcharTable['7'] = 7;
    _pfcharTable['8'] = 8;
    _pfcharTable['9'] = 9;
    _pfcharTable['A'] = 10; _pfcharTable['a'] = 10;
    _pfcharTable['B'] = 11; _pfcharTable['b'] = 11;
    _pfcharTable['C'] = 12; _pfcharTable['c'] = 12;
    _pfcharTable['D'] = 13; _pfcharTable['d'] = 13;
    _pfcharTable['E'] = 14; _pfcharTable['e'] = 14;
    _pfcharTable['F'] = 15; _pfcharTable['f'] = 15;
}
