/******************************************************************************

   Copyright 1991 Lee Iverson

   Permission to use, copy, modify, and distribute this software and its
   documentation for any purpose without fee is hereby granted without fee,
   provided that the above copyright notice appear in all copies and that
   both that copyright notice and this permission notice appear in
   supporting documentation, and that the name of McGill not be used in
   advertising or publicity pertaining to distribution of the software
   without specific, written prior permission.  McGill makes no
   representations about the suitability of this software for any purpose.
   It is provided "as is" without express or implied warranty.

   MCGILL DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL
   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL MCGILL
   BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
   WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION
   OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
   CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

   Author:
     Lee Iverson <leei@mcrcim.mcgill.edu>,
     McGill Research Centre for Intelligent Machines (McRCIM)

******************************************************************************/

static char RCSid[] = "$Id: pbm_read.c,v 0.9.1.1 1991/11/06 00:09:25 leei Exp $";

/* From pbm.c:
 *
 * portable bit map (pbm) format images
 *
 * jim frost 09.27.89
 *
 * patched by W. David Higgins (wdh@mkt.csd.harris.com) to support
 * raw-format PBM files.
 *
 * patched by Ian MacPhedran (macphed@dvinci.usask.ca) to support
 * PGM and PPM files (03-July-1990)
 */

#include <vim/proto.h>
#include <vim/pFile.h>

#include <vim/BitImageP.h>

#define NOTINT  -1
#define COMMENT -2
#define SPACE   -3
#define NEWLINE -4

#define BADREAD    0 /* read error */
#define NOTPBM     1 /* not a pbm file */
#define PBMNORMAL  2 /* pbm normal type file */
#define PBMCOMPACT 3 /* pbm compacty type file */
#define PBMRAWBITS 4 /* pbm raw bits type file */
#define PGMNORMAL  5 /* pgm normal type file */
#define PGMRAWBITS 6 /* pgm raw bytes type file */
#define PPMNORMAL  7 /* ppm normal type file */
#define PPMRAWBITS 8 /* ppm raw bytes type file */

GenericImage
pbm_read _IMPL((in, ident), pFILE in _AND char *ident )
{
    BitImage im;
    char buf[3];
    int pbm_type;
    int src;
    unsigned int  linelen;
    byte          srcmask, destmask;
    byte         *destptr, *destline;
    int numbytes, numread;
    int x, y;
    declare_BitImage(pbm);

    /* Determine the type of the PBM image. */
    buf[0] = pfgetc(in);
    buf[1] = pfgetc(in);
    buf[2] = '\0';
    pbm_type = 
      (!strncmp("P1",buf,2))? PBMNORMAL:
      (!strncmp("P4",buf,2))? PBMRAWBITS: NOTPBM;

    pfgetuhint(&pbm_width,in);
    pfgetuhint(&pbm_height,in);
    pfgets(NULL,-1,in);
    
    im = (BitImage) 
      ImCreateImage( ident, bitImageClass, pbm_width, pbm_height, NULL, 0 );
    access_BitImage(im,pbm);

    switch ( pbm_type ) {
      case PBMRAWBITS:
	destline= pbm_data;
	linelen= (pbm_width + 7) / 8;
	numbytes= linelen * pbm_height;
	srcmask= 0;		/* force initial read */
	numread= 0;
	for (y= 0; y < pbm_height; y++) {
	    destptr= destline;
	    destmask= 0x80;
	    if (srcmask != 0x80) {
		srcmask= 0x80;
		if ((numread < numbytes) && ((src = pfgetc(in)) == EOF)) {
		    fprintf(stderr,"%s: Short image\n", ident);
		    goto done;
		}
		numread++;
	    }
	    for (x= 0; x < pbm_width; x++) {
		if (src & srcmask)
		  *destptr |= destmask;
		if (! (destmask >>= 1)) {
		    destmask= 0x80;
		    destptr++;
		}
		if (! (srcmask >>= 1)) {
		    srcmask= 0x80;
		    if ((numread < numbytes) && ((src= pfgetc(in)) == EOF)) {
			fprintf(stderr,"%s: Short image\n", ident);
			goto done;
		    }
		    numread++;
		}
	    }
	    destline += linelen;
	}
	break;
      case PBMNORMAL:
	linelen = (pbm_width / 8) + (pbm_width % 8 ? 1 : 0);
	destline = pbm_data;
	for (y= 0; y < pbm_height; y++) {
	    destptr= destline;
	    destmask= 0x80;
	    for (x= 0; x < pbm_width; x++) {
		if ((src = pfnextc(in)) < 0) {
		    fprintf(stderr,"%s: Short image\n", ident);
		    goto done;
		}
		if ( pfcharval(src) < 0 ) {
		    fprintf(stderr,"%s: Bad image data\n", ident);
		    goto done;
		}
		
		switch (pfcharval(src)) {
		  case 1:
		    *destptr |= destmask;
		  case 0:
		    if (! (destmask >>= 1)) {
			destmask= 0x80;
			destptr++;
		    }
		    break;
		  default:
		    fprintf(stderr,"%s: Bad image data\n", ident);
		    goto done;
		}
	    }
	    destline += linelen;
	}
	break;
      default:
	fprintf(stderr,"%s: Unrecognized PBM format\n",ident);
	ImDestroyImage((GenericImage)im);
	return 0;
    }

  done:
    end_BitImage(im,pbm);
    return (GenericImage)im;
}
