/******************************************************************************

   Copyright 1991 Lee Iverson

   Permission to use, copy, modify, and distribute this software and its
   documentation for any purpose without fee is hereby granted without fee,
   provided that the above copyright notice appear in all copies and that
   both that copyright notice and this permission notice appear in
   supporting documentation, and that the name of McGill not be used in
   advertising or publicity pertaining to distribution of the software
   without specific, written prior permission.  McGill makes no
   representations about the suitability of this software for any purpose.
   It is provided "as is" without express or implied warranty.

   MCGILL DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL
   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL MCGILL
   BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
   WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION
   OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
   CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

   Author:
     Lee Iverson <leei@mcrcim.mcgill.edu>,
     McGill Research Centre for Intelligent Machines (McRCIM)

******************************************************************************/

static char RCSid[] = "$Id: readImage.c,v 0.9.1.5 1991/12/19 19:47:10 leei Exp $";

#include <string.h>
#include <stdio.h>
#include <stdlib.h>

#include <sys/types.h>
#include <sys/stat.h>

#include <unistd.h>
#include <ctype.h>

#include "config.h"
#include "GenImageP.h"
#include "regexp.h"

#include "proto.h"
#include "debug.h"
#include "pFile.h"
#include "strstr.h"

#ifdef DLD
#include <dld.h>
#endif

#include "pbm.h"
#include "pgm.h"
#include "ppm.h"
#include "gif.h"
#include "gim_util.h"

#ifdef RLE
#include "rle_read.h"
#endif	/* RLE */

typedef int (*ident_fp) __((pFILE in));
typedef GenericImage (*read_fp) __((pFILE in, char *id));
typedef void (*write_fp) __((pFILE in));

typedef struct _image_type {
    char *name;
    char *file_pattern;
    int (*ident_f) __((pFILE in));
    GenericImage (*read_f) __((pFILE in, char *id));
    void (*write_f) __((pFILE in));
    char *filter;
} image_type;

char *exclude_file_pattern = "([Mm]akefile|\\.(cc|ad|[chylpo]))$";

#ifdef BTOA
int btoa_ident __((pFILE in ));
#endif

#ifdef COMPRESS
int compress_ident __(( pFILE in ));
#endif

#ifdef YABBA
int yabba_ident __(( pFILE in ));
int whap_ident __(( pFILE in ));
#endif

#ifdef JPEG
char *jpeg_file_pattern   = ".jp(|e)g$";
#endif

#ifdef DLD
char *gim_file_pattern    = "\\.(gim|xdr)$";
#endif

#ifdef PBMPLUS
char *xbm_file_pattern    = "\\.(x|xbm(|10|11)|bitmap)$";
char *rast_file_pattern   = "\\.(r|rast)$";
char *tiff_file_pattern   = "\\.tif(|f)$";
char *macp_file_pattern   = "\\.mac(|p)$";
char *g3_file_pattern     = "\\.(g3|fax)$";
char *xwd_file_pattern    = "\\.x(|10|11)wd$";
char *brush_file_pattern  = "\\.brush$";
char *gem_file_pattern    = "\\.img$";
char *pcx_file_pattern    = "\\.pcx$";
char *pict_file_pattern   = "\\.pic(|t|t2)$";
char *fs_file_pattern     = "\\.(fs|face)$";
char *hips_file_pattern   = "\\.hips$";
char *fits_file_pattern   = "\\.fits$";
char *ilbm_file_pattern   = "\\.(iff|ilbm)$";
char *lispm_file_pattern  = "\\.lispm$";
char *mtv_file_pattern    = "\\.mtv$";
char *qrt_file_pattern    = "\\.qrt$";
char *tga_file_pattern    = "\\.(tga|targa)$";
char *xim_file_pattern    = "\\.xim$";
char *xpm_file_pattern    = "\\.xpm(|2|3)$";
char *pi1_file_pattern    = "\\.pi1$";
char *pi3_file_pattern    = "\\.pi3$";
char *spu_file_pattern    = "\\.spu$";
char *spc_file_pattern    = "\\.spc$";
char *ybm_file_pattern    = "\\.(ybm|face)$";
char *cmuwm_file_pattern  = "\\.cmuwm$";
char *mgr_file_pattern    = "\\.mgr$";

int cmuwm_ident __(( pFILE in ));
int ilbm_ident __(( pFILE in ));
int lispm_ident __(( pFILE in ));
int mgr_ident __(( pFILE in ));
int macp_ident __(( pFILE in ));
int rast_ident __(( pFILE in ));
int tiff_ident __(( pFILE in ));
int ybm_ident __(( pFILE in ));
int xpm_ident __(( pFILE in ));
int xbm_ident __(( pFILE in ));
#endif

static image_type init_image_types[] = {
#ifdef COMPRESS
    { "compress", NULL, compress_ident,	NULL, NULL, "uncompress -c" },
#endif
#ifdef YABBA
    { "yabba", NULL, yabba_ident,	NULL, NULL, "unyabba" },
    { "whap", NULL, whap_ident,	NULL, NULL, "unwhap" },
#endif
#ifdef BTOA
    { "btoa",  NULL, btoa_ident, NULL, NULL, "atob" },
#endif
#ifdef JPEG
    { "jpeg",  (char*)&jpeg_file_pattern,  NULL, NULL, NULL, "djpeg" },
#endif
#ifdef PBMPLUS
    { "brush", (char*)&brush_file_pattern, NULL, NULL, NULL, "brushtopbm" },
    { "cmuwm", (char*)&cmuwm_file_pattern, cmuwm_ident, 
	NULL, NULL, "cmuwmtopbm" },
    { "fs",    (char*)&fs_file_pattern,    NULL, NULL, NULL, "fstopgm" },
    { "fits",  (char*)&fits_file_pattern,  NULL, NULL, NULL, "fitstopgm" },
    { "g3",    (char*)&g3_file_pattern,    NULL, NULL, NULL, "g3topbm" },
    { "gem",   (char*)&gem_file_pattern,   NULL, NULL, NULL, "gemtopbm" },
    { "hips",  (char*)&hips_file_pattern,  NULL, NULL, NULL, "hipstopgm" },
    { "ilbm",  (char*)&ilbm_file_pattern,  ilbm_ident, 
	NULL, NULL, "ilbmtoppm" },
    { "lispm", (char*)&lispm_file_pattern, lispm_ident, 
	NULL, NULL, "lispmtopgm" },
    { "macp",  (char*)&macp_file_pattern,  macp_ident,
	NULL, NULL, "macptopbm" },
    { "mgr",   (char*)&mgr_file_pattern,   mgr_ident, 
	NULL, NULL, "mgrtopbm" },
    { "mtv",   (char*)&mtv_file_pattern,   NULL, NULL, NULL, "mtvtoppm" },
    { "pcx",   (char*)&pcx_file_pattern,   NULL, NULL, NULL, "pcxtoppm" },
    { "pi1",   (char*)&pi1_file_pattern,   NULL, NULL, NULL, "pi1toppm" },
    { "pi3",   (char*)&pi3_file_pattern,   NULL, NULL, NULL, "pi3topbm" },
    { "pict",  (char*)&pict_file_pattern,  NULL, NULL, NULL, "picttoppm" },
    { "qrt",   (char*)&qrt_file_pattern,   NULL, NULL, NULL, "qrttoppm" },
    { "rast",  (char*)&rast_file_pattern,  rast_ident, 
	NULL, NULL, "rasttopnm" },
    { "spc",   (char*)&spc_file_pattern,   NULL, NULL, NULL, "spctoppm" },
    { "spu",   (char*)&spu_file_pattern,   NULL, NULL, NULL, "sputoppm" },
    { "tga",   (char*)&tga_file_pattern,   NULL, NULL, NULL, "tgatoppm" },
    { "tiff",  (char*)&tiff_file_pattern,  tiff_ident,
	NULL, NULL, "tifftopnm" },
    { "xbm",   (char*)&xbm_file_pattern,   xbm_ident, 
	NULL, NULL, "xbmtopbm" },
    { "xim",   (char*)&xim_file_pattern,   NULL, NULL, NULL, "ximtoppm" },
    { "xpm",   (char*)&xpm_file_pattern,   xpm_ident, 
	NULL, NULL, "xpmtoppm" },
    { "xwd",   (char*)&xwd_file_pattern,  NULL, NULL, NULL, "xwdtopnm" },
    { "ybm",   (char*)&ybm_file_pattern,   ybm_ident,
	NULL, NULL, "ybmtoppm" },
#endif
#if defined(DLD)
    { "gim", (char*)&gim_file_pattern, NULL, NULL, NULL },
#endif
#if defined(DLD)
    { "pbm", (char*)&pbm_file_pattern, pbm_ident, NULL, NULL },
    { "pgm", (char*)&pgm_file_pattern, pgm_ident, NULL, NULL },
    { "ppm", (char*)&ppm_file_pattern, ppm_ident, NULL, NULL },
#ifdef RLE
    { "rle", (char*)&rle_file_pattern, rle_ident, NULL, NULL },
#endif		/* RLE */
    { "gif", (char*)&gif_file_pattern, gif_ident, NULL, NULL }
#else
    { "pbm", (char*)&pbm_file_pattern, pbm_ident, pbm_read, NULL },
    { "pgm", (char*)&pgm_file_pattern, pgm_ident, pgm_read, NULL },
    { "ppm", (char*)&ppm_file_pattern, ppm_ident, ppm_read, NULL },
#ifdef RLE
    { "rle", (char*)&rle_file_pattern, rle_ident, rle_read, NULL },
#endif		/* RLE */
    { "gif", (char*)&gif_file_pattern, gif_ident, gif_read, NULL }
#endif /* DLD */
};

static image_type *image_types = init_image_types;
static Boolean image_types_freeable = False;
static int num_image_types = sizeof(init_image_types)/sizeof(image_type);
static int last_image_type = sizeof(init_image_types)/sizeof(image_type);

static void
Initialize _IMPL_VOID
{
    static int initialized = 0;
    int i;

    if ( !initialized ) {
	exclude_file_pattern = (char *) regcomp(exclude_file_pattern);

	for ( i = 0; i < num_image_types; ++i ) {
	    if ( image_types[i].file_pattern ) {
		char buf[128];
		image_types[i].file_pattern =
		  *(char**)(image_types[i].file_pattern);
		strcpy(buf, image_types[i].file_pattern);
#ifdef COMPRESS
		if ( buf[strlen(buf)-1] == '$' ) {
		    /* Add an optional compression id on to the end. */
		    strcpy(buf + strlen(buf) - 1, "(|\\.Z)$");
		}
#endif
#ifdef YABBA
		if ( buf[strlen(buf)-1] == '$' ) {
		    /* Add an optional yabba id on to the end. */
		    strcpy(buf + strlen(buf) - 1, "(|\\.Y)$");
		}
		if ( buf[strlen(buf)-1] == '$' ) {
		    /* Add an optional whap id on to the end. */
		    strcpy(buf + strlen(buf) - 1, "(|\\.AP)$");
		}
#endif
#ifdef BTOA
		if ( buf[strlen(buf)-1] == '$' ) {
		    /* Add an optional btoa id on to the end. */
		    strcpy(buf + strlen(buf) - 1, "(|\\.btoa)$");
		}
#endif
		if ( buf[strlen(buf)-1] == '$' ) {
		    /* Recognize GNU version control extensions. */
		    strcpy(buf + strlen(buf) - 1, "(|~|\\.[0-9]+~)$");
		}
		image_types[i].file_pattern = (char *) regcomp(buf);
	    }
	}

	initialized = 1;
    }
}

static GenericImage 
  readImagePF __(( pFILE in, char *fname, image_type *type ));

static int type_seq[16];
static int ntype;

GenericImage
readImage _IMPL((fname), char *fname )
{
    pFILE in;
    GenericImage im;
    ImFileAssoc assoc = ImFindFile(fname);

    if ( assoc ) {
	struct stat sbuf;
	if ( stat(fname, &sbuf) < 0 ) {
	    fprintf(stderr, "File %s has been deleted!\n", fname);
	    return 0;
	}
	if ( sbuf.st_mtime <= assoc->image->generic.timestamp )
	  return assoc->image;
    }

    Initialize();

    
    /* Open the file and read in header information. */
    if ( !(in = pfopen( fname )) ) {
	fprintf(stderr,"%s: Cannot open.\n", fname );
	return 0;
    }

    ntype = 0;
    im = readImagePF(in, fname, NULL);
    pfclose(in);

    return im;
}

static char *strtolower __(( char *str ));
static char fnamelow[512];

#ifdef DLD
typedef void *(*pfvp)();
typedef void (*pfv)();

static image_type *NewType __(( char *name ));
pfvp dld_link_until_execable __(( char *func_name, char *fname ));
#endif /* DLD */

static GenericImage
readImagePF _IMPL((in, fname, prev_type),
		  pFILE in _AND char *fname _AND image_type *prev_type )
{
    GenericImage im = 0;
    image_type *type = 0;
    char ident[128];
    int i;
    char *p;
    static int nuids = 0;
    static int uid[64];

    /* Read descriptor and identify image type */
    for ( i = 0, type = image_types; i < num_image_types; ++i, ++type ) {
	if ( type->ident_f && (*type->ident_f)(in) )
	  break;
    }

    /* Check file-defined type. */
    if ( i == num_image_types && (p = gim_type(in)) ) {
	for ( i = 0, type = image_types; i < num_image_types; ++i, ++type ) {
	    if ( !strcmp(type->name, p) )
	      break;
	}
#ifdef DLD
	if ( i == num_image_types )
	  type = NewType(p);
	else
#endif /* DLD */
	  free(p);
    }

    /* Failed using magic number/header information, so try file patterns. */
    if ( i == num_image_types ) {
	strcpy( fnamelow, fname ); strtolower(fnamelow);
	for ( i = 0, type = image_types; i < num_image_types; ++i, ++type ) {
	    if ( type->file_pattern &&
		 (regexec((regexp *)(type->file_pattern),fname) ||
		  regexec((regexp *)(type->file_pattern),fnamelow)) )
	      break;
	}
    }

    /* Make sure that we actually recognized it. */
    if ( i == num_image_types ) {
	fprintf(stderr,"%s: Unrecognized image type.\n",fname);
	return 0;
    }
    if ( type == prev_type ) {
	fprintf(stderr,"%s: Cannot read image type %s.\n", fname, type->name);
	return 0;
    }
    DEBUG((DBG_LD, "Recognized image %s: type %s\n", fname, type->name));

    /* Remember the type at this point in the sequence. */
    type_seq[ntype++] = type - image_types;

    if ( type->filter ) {
	int uidx = nuids;
	int rtof[2], ftor[2];
	char *filter = type->filter;

	/* Check to see if environment defines an override for filter. */
	{
	    char envbuf[128];
	    sprintf(envbuf, "VIMAGE_%s", type->name);
	    if ( getenv(envbuf) ) filter = getenv(envbuf);
	}

	nuids += 2;
	uid[uidx] = 0;
	uid[uidx+1] = 0;

	if ( pipe(rtof) || pipe(ftor) ) {
	    fprintf(stderr, "Failed to create pipes for %s\n", type->name);
	    return 0;
	}

	DEBUG((DBG_LD, "Start filter \"%s\"\n", filter));
	if ( !(uid[uidx+1] = vfork()) ) {
	    /* This process is the filter. */
	    char buf[256];

	    close(ftor[0]); close(rtof[1]);
	    if ( rtof[0] != 0 ) { dup2(rtof[0],0); close(rtof[0]); }
	    if ( ftor[1] != 1 ) { dup2(ftor[1],1); close(ftor[1]); }
	    strcpy( buf, "exec " ); strcat( buf, filter );
	    if ( execl( "/bin/sh", "sh", "-c", buf, NULL ) ) 
	      _exit(127);
	}

	/* Must use fork() here to prevent deadlock. */
	if ( !(uid[uidx] = fork()) ) {
	    /* This process writes the stream out to the filter. */
	    int n; char *buf;

	    close(rtof[0]); close(ftor[0]); close(ftor[1]);
#ifdef BSD_STDIO
	    n = pfpushback(&buf, in);
	    if ( n ) write(rtof[1], buf, n);
	    if ( in->fp->_cnt > 0 ) write(rtof[1], in->fp->_ptr, in->fp->_cnt);
	    dup2(fileno(in->fp), 0); fclose(in->fp);
	    dup2(rtof[1], 1); close(rtof[1]);
	    if ( execl( "/bin/cat", "cat", NULL ) ) 
	      _exit(127);
#else
	    buf = (char *) alloca(BUFSIZ);
	    while ( n = pfread(buf,1,BUFSIZ,in) ) write(rtof[1],buf,n);
	    close(rtof[1]);
	    _exit(0);
#endif
	}

	/* This is the main process which reads the filter output. */
	close(rtof[0]); close(rtof[1]); close(ftor[1]);
	{
	    pFILE new = pfreopen(fdopen(ftor[0],"r"));
	    im = readImagePF(new, fname, type);
	    pfclose(new);
	}

	/* Wait for both sub processes to exit. */
	while ( uid[uidx] != 0 || uid[uidx+1] != 0 ) {
	    int status;
	    int wuid = wait(&status);
	    if ( wuid < 0 ) {
		perror(type->name);
		break;
	    }
	    for ( i = 0; i < nuids; ++i ) 
	      if ( wuid == uid[i] ) uid[i] = 0;
	}
	nuids -= 2;
	return im;
    }

#ifdef DLD
    if ( !type->read_f ) {
	char *func_name = XtMalloc( strlen(type->name) + 6 );
	sprintf(func_name, "%s_read", type->name);
	type->read_f = (read_fp) dld_link_until_execable(func_name, fname);
	XtFree(func_name);
    }
#endif /* DLD */

    if ( type->read_f ) {
	/* Create an identifier. */
	strcpy(ident,(p = rindex(fname,'/'))? p+1: fname);

	/* Actually read it using the specific command. */
	im = (*type->read_f)(in,ident);
    } else {
	fprintf(stderr, "No reader function for file type %s\n", type->name);
    }

    /* Update file-specific information. */
    if ( im ) {
	im->generic.file_name = strdup(fname);
	im->generic.file_type = type->name;
	time(&(im->generic.timestamp));
	im->generic.changed = 1;
	ImAddFile(im);
    }

    return im;
}

#ifdef DLD
static image_type *
NewType _IMPL((name), char *name )
{
    image_type *new;

    /* Maybe extend the table. */
    if ( num_image_types == last_image_type ) {
	image_type *prev = image_types;
	last_image_type += 10;
	image_types = 
	  (image_type *) XtMalloc(last_image_type * sizeof(image_type));
	bzero(image_types, last_image_type *  sizeof(image_type));
	bcopy(prev, image_types, num_image_types *  sizeof(image_type));
	if ( image_types_freeable ) XtFree(prev);
	image_types_freeable = True;
    }
    
    new = &image_types[num_image_types++];
    new->name = name;
    return new;
}

static char **load_fnames = 0;
static int num_fnames = 0;
static int last_fname = 0;

static char **
new_load_fname _IMPL_VOID
{
    if ( num_fnames == last_fname ) {
	last_fname += 10;
	load_fnames = 
	  (char **) XtRealloc(load_fnames, last_fname * sizeof(char*));
    }

    return &load_fnames[num_fnames++];
}

void
dld_resolve_directives _IMPL((s), char *s)
{
    char *p, *q;
    char buf[BUFSIZ];

    strcpy(buf, s);
    for ( p = buf, q = s; *p; ++p ) {
	if ( *p == '%' ) {
	    switch( *++p ) {
	      case 'A':
	      case 'a':
		strcpy(q, ARCH);
		q = index(q,'\0');
		break;
	      case 'H':
	      case 'h':
		strcpy(q, getenv("HOME"));
		q = index(q,'\0');
		break;
	      default:
		*q++ = *p;
		break;
	    }
	} else {
	    *q++ = *p;
	}
    }
    *q = '\0';
}

void
dld_init_load_files _IMPL((path, files), char *path _AND char *files )
{
    char *p, *q, *r, *s;
    char **fnamep;
    int path_len = strlen(path);
    int files_len = strlen(files);
    struct stat sbuf;
    char buf[BUFSIZ];
    Boolean found;

    for ( r = files; r <= files + files_len; r = s+1 ) {
	found = False;

	while ( isspace(*r) ) ++r;
	s = r;
	while ( !isspace(*s) ) ++s;
	if ( *r == '/' ) {
	    sprintf(buf, "%.*s", s-r, r);
	    if ( !found && !stat(buf, &sbuf) ) {
		fnamep = new_load_fname();
		*fnamep = strdup(buf);
		found = True;
	    }
	} else {
	    for ( p = path; p <= path + path_len; p = q+1 ) {
		(q = index(p,':')) || (q = index(p,'\0'));
		if ( !strncmp(r,"-l",2) )
		  sprintf(buf, "%.*s/lib%.*s.a", q-p, p, s-r-2, r+2);
		else
		  sprintf(buf, "%.*s/%.*s", q-p, p, s-r, r);
		dld_resolve_directives(buf);
		if ( *buf != '/' || (!found && !stat(buf, &sbuf)) ) {
		    fnamep = new_load_fname();
		    *fnamep = strdup(buf);
		    found |= (*buf == '/');
		}
	    }
	}
	if ( found )
	  DEBUG((DBG_LD, "DLD: %.*s => %s\n", s-r, r, *fnamep));
    }
}

pfvp
dld_link_until_execable _IMPL((func_name, relative_to),
			      char *func_name _AND char *relative_to )
{
    int i, j;
    char **fnamep;
    char **undefp;
    Boolean found;
    struct stat sbuf;

    dld_create_reference(func_name);
    DEBUG((DBG_LD, "DLD: Resolve reference %s\n", func_name));
    for ( i = 0, fnamep = load_fnames; 
	  i < num_fnames && !dld_function_executable_p(func_name); 
	  ++i, ++fnamep ) {
	found = False;

#if DEBUGGING
	DEBUG((DBG_LD, "DLD:  Unresolved"));
	undefp = dld_list_undefined_sym();
	for ( j = 0; j < dld_undefined_sym_count; ++j )
	  DEBUG((DBG_LD, " %s", *undefp++));
	DEBUG((DBG_LD, "\n"));
#endif

	/* Relative pathnames should examine file path below filename. */
	if ( relative_to && !dld_function_executable_p(func_name) &&
	     **fnamep != '/' ) {
	    char buf[BUFSIZ];
	    buf[0] = '\0';
	    if ( *relative_to != '/' ) {
		getwd(buf);
		strcat(buf, "/");
	    }
	    strcat(buf, relative_to);
	    strrchr(buf, '/')[1] = '\0';
	    strcat(buf, *fnamep);
	    if ( !stat(buf, &sbuf) ) {
		dld_link(buf); DEBUG((DBG_LD, "DLD: link %s\n", buf));
		found = True;
	    }
	}
	if ( !found && !stat(*fnamep, &sbuf) ) {
	    dld_link(*fnamep); DEBUG((DBG_LD, "DLD: link %s\n", *fnamep));
	    found = True;
	}
    }

    if ( dld_function_executable_p(func_name) ) {
	return (pfvp) dld_get_func(func_name);
    } else {
	fprintf(stderr, "Attempt to load %s failed.\n  Undefined:", func_name);
	undefp = dld_list_undefined_sym();
	for ( j = 0; j < dld_undefined_sym_count; ++j )
	  fprintf(stderr, " %s", *undefp++);
	fprintf(stderr, "\n");
	return 0;
    }
}
#endif /* DLD */

Boolean
acceptImageFile _IMPL((fname), String fname)
{
    int i;
    image_type *type;

    Initialize();

    /* Make a lowercase copy of the string. */
    strcpy(fnamelow,fname); strtolower(fnamelow);

    if ( regexec((regexp *)exclude_file_pattern, fname) ) return False;

    for ( i = 0, type = image_types; i < num_image_types; ++i, ++type ) {
	if ( type->file_pattern && 
	     (regexec((regexp *)(type->file_pattern),fname) ||
	      regexec((regexp *)(type->file_pattern),fnamelow)) )
	  return True;
    }
    return False;
}

static char *
strtolower _IMPL((str), char *str )
{
    char *p;

    for ( p = str; *p; ++p ) 
      if ( isupper(*p) ) *p = tolower(*p);
    return str;
}

#ifdef BTOA
int
btoa_ident _IMPL((in), pFILE in )
{
    char head[5];
    if ( pfpeek(head,1,5,in) != 5 ) return False;
    return !strncmp("xbtoa",head,5);
}
#endif

#ifdef COMPRESS
int
compress_ident _IMPL((in), pFILE in )
{
    unsigned char head[2];
    if ( pfpeek((char *)head,1,2,in) != 2 ) return False;
    return (head[0] == 0x1f) && (head[1] == 0x9d);
}
#endif

#ifdef YABBA
int
yabba_ident _IMPL((in), pFILE in )
{
    unsigned char head[2];
    if ( pfpeek((char *)head,1,5,in) != 5 ) return False;
    /* Header: ^Y^A^B^B^A */
    return !strncmp("\031\001\002\002\001",head,5);
}

int
whap_ident _IMPL((in), pFILE in )
{
    unsigned char head[2];
    if ( pfpeek((char *)head,1,4,in) != 4 ) return False;
    /* Header: ^W^H^A^P */
    return !strncmp("\027\010\001\020",head,4);
}
#endif

#ifdef PBMPLUS
int
cmuwm_ident _IMPL((in), pFILE in )
{
    char head[4];
    if ( pfpeek(head,1,4,in) != 4 ) return False;
    return !strncmp("\361\0\100\273",head,4);
}

int
rast_ident _IMPL((in), pFILE in )
{
    char head[4];
    if ( pfpeek(head,1,4,in) != 4 ) return False;
    return !strncmp("\131\246\152\225",head,4);
}

int
mgr_ident _IMPL((in), pFILE in )
{
    char head[2];
    if ( pfpeek(head,1,2,in) != 2 ) return False;
    return !strncmp("yz",head,2);
}

int
tiff_ident _IMPL((in), pFILE in )
{
    char head[2];
    if ( pfpeek(head,1,2,in) != 2 ) return False;
    return
      !strncmp("\115\115",head,2) ||
      !strncmp("\111\111",head,2);
}

int
ilbm_ident _IMPL((in), pFILE in )
{
    char head[4];
    if ( pfpeek(head,1,4,in) != 4 ) return False;
    return !strncmp("ILBM",head,4);
}

int
lispm_ident _IMPL((in), pFILE in )
{
    char head[21];
    if ( pfpeek(head,1,21,in) != 21 ) return False;
    return !strncmp("This\040is\040a\040BitMap\040file",head,21);
}

int
ybm_ident _IMPL((in), pFILE in )
{
    char head[2];
    if ( pfpeek(head,1,2,in) != 2 ) return False;
    return !strncmp("!!",head,2);
}

int
macp_ident _IMPL((in), pFILE in )
{
    char head[69];
    if ( pfpeek(head,1,69,in) != 69 ) return False;
    return !strncmp("PNTG",head+65,4);
}

int
xbm_ident _IMPL((in), pFILE in )
{
    char head[8];
    if ( pfpeek(head,1,8,in) != 8 ) return False;
    return !strncmp("#define ",head,8);
}

int
xpm_ident _IMPL((in), pFILE in )
{
    char head[6];
    if ( pfpeek(head,1,6,in) != 6 ) return False;
    return !strncmp("/* XPM",head,6);
}

#endif
